@@ -1,15 +1,21 @@
 """
     Binding classes.
 """
+from __future__ import annotations
 
 # standard libraries
+import functools
 import typing
+import weakref
 
 # third party libraries
 # none
 
 # local libraries
-# none
+from . import Converter
+from . import Observable
+from . import Validator
+from .ReferenceCounting import weak_partial
 
 
 class Binding:
@@ -39,64 +45,59 @@ class Binding:
     cases.
     """
 
-    def __init__(self, source, *, converter=None, validator=None, fallback=None):
+    def __init__(self, source: typing.Optional[Observable.ObservableLike], *,
+                 converter: typing.Optional[Converter.ConverterLike[typing.Any, typing.Any]] = None,
+                 validator: typing.Optional[Validator.ValidatorLike[typing.Any]] = None,
+                 fallback: typing.Optional[typing.Any] = None) -> None:
         super().__init__()
         self.__converter = converter
         self.__validator = validator
         self.fallback = fallback
-        self.source_getter = None
-        self.source_setter = None
-        self.target_setter = None
-        self.__source = source
+        self.source_getter: typing.Optional[typing.Callable[[], typing.Any]] = None
+        self.source_setter: typing.Optional[typing.Callable[[typing.Any], None]] = None
+        self.target_setter: typing.Optional[typing.Callable[[typing.Any], None]] = None
+        # Python 3.9+: should weakref.ReferenceType
+        self.__source_ref: typing.Optional[typing.Any] = weakref.ref(source) if source else None
         self._closed = False
 
     # not thread safe
-    def close(self):
-        """Close the binding.
-
-        Closes the binding. Subclasses can use this to perform any shutdown
-        related actions. Not thread safe.
-        """
-        self.__source = None
-        self.source_getter = None
-        self.source_setter = None
-        self.target_setter = None
-        self._closed = True
+    def close(self) -> None:
+        pass
 
     @property
-    def source(self):
+    def source(self) -> typing.Optional[typing.Any]:
         """Return the source of the binding. Thread safe."""
-        return self.__source
+        return self.__source_ref() if self.__source_ref else None
 
     @property
-    def converter(self):
+    def converter(self) -> typing.Optional[Converter.ConverterLike[typing.Any, typing.Any]]:
         """Return the converter (from source to target). Thread safe."""
         return self.__converter
 
     @property
-    def validator(self):
+    def validator(self) -> typing.Optional[Validator.ValidatorLike[typing.Any]]:
         """Return the validator (of converted value). Thread safe."""
         return self.__validator
 
     # thread safe
-    def __back_converted_value(self, target_value):
+    def __back_converted_value(self, target_value: typing.Optional[typing.Any]) -> typing.Optional[typing.Any]:
         """Return the back converted value (from target to source). Thread safe."""
         return self.__converter.convert_back(target_value) if self.__converter else target_value
 
     # thread safe
-    def __converted_value(self, source_value):
+    def __converted_value(self, source_value: typing.Optional[typing.Any]) -> typing.Optional[typing.Any]:
         """Return the converted value (from source to target). Thread safe."""
         return self.__converter.convert(source_value) if self.__converter else source_value
 
     # thread safe
-    def __validated_value(self, source_value):
+    def __validated_value(self, source_value: typing.Optional[typing.Any]) -> typing.Optional[typing.Any]:
         """Return the converted value (from source to target). Thread safe."""
         return self.__validator.validate(source_value) if self.__validator else source_value
 
     # public methods. subclasses must make sure these methods work as expected.
 
     # thread safe
-    def update_source(self, target_value):
+    def update_source(self, target_value: typing.Optional[typing.Any]) -> None:
         """Update source with back converted target value.
 
         Update the source from the target value. The target value will be back converted.
@@ -106,12 +107,12 @@ class Binding:
 
         Thread safe.
         """
-        if self.source_setter:
+        if callable(self.source_setter):
             converted_value = self.__validated_value(self.__back_converted_value(target_value))
             self.source_setter(converted_value)
 
     # not thread safe
-    def update_target(self, source_value):
+    def update_target(self, source_value: typing.Optional[typing.Any]) -> None:
         """Update target with converted source value.
 
         Call the target setter with the unconverted value from the source.
@@ -126,7 +127,7 @@ class Binding:
         self.update_target_direct(self.__converted_value(source_value))
 
     # not thread safe
-    def update_target_direct(self, converted_value):
+    def update_target_direct(self, converted_value: typing.Optional[typing.Any]) -> None:
         """Update target directly with converted value.
 
         Call the target setter with the already converted value.
@@ -138,11 +139,11 @@ class Binding:
 
         Not thread safe.
         """
-        if self.target_setter:
+        if callable(self.target_setter):
             self.target_setter(converted_value)
 
     # thread safe
-    def get_target_value(self):
+    def get_target_value(self) -> typing.Optional[typing.Any]:
         """Return target value by converting source.
 
         Get the value from the source that will be set on the target.
@@ -152,64 +153,13 @@ class Binding:
 
         Thread safe.
         """
-        if self.source_getter:
+        if callable(self.source_getter):
             source = self.source_getter()
             if source is not None:
                 return self.__converted_value(source)
         return self.fallback
 
 
-class ListBinding(Binding):
-    """One way binding from a source list to a target.
-
-    Observes item inserted and item removed events on a source and if the event
-    matches the key, calls the functions assigned to inserter and remover fields.
-
-    Source list must support item_inserted_event and item_removed_event with the
-    signatures:
-        item_inserted(key: str, item: Any, before_index: int) -> None
-        item_removed(key: str, item: Any, index: int) -> None
-
-    Client should set the inserter and remover functions to callables with the
-    signatures:
-        inserter(item: Any, before_index: int) -> None
-        remover(index: int) -> None.
-
-    The owner should call close on this object.
-    """
-
-    def __init__(self, source, key: str):
-        super().__init__(source)
-        self.__key = key
-        self.inserter: typing.Optional[typing.Callable[[typing.Any, int], None]] = None
-        self.remover: typing.Optional[typing.Callable[[int], None]] = None
-
-        # thread safe
-        def item_inserted(key_: str, item: typing.Any, before_index: int) -> None:
-            if key_ == self.__key and callable(self.inserter):
-                self.inserter(item, before_index)
-
-        # thread safe
-        def item_removed(key_: str, item: typing.Any, index: int) -> None:
-            if key_ == self.__key and callable(self.remover):
-                self.remover(index)
-
-        self.__item_inserted_listener = source.item_inserted_event.listen(item_inserted)
-        self.__item_removed_listener = source.item_removed_event.listen(item_removed)
-
-    def close(self):
-        self.__item_inserted_listener.close()
-        self.__item_inserted_listener = None
-        self.__item_removed_listener.close()
-        self.__item_removed_listener = None
-        super().close()
-
-    @property
-    def items(self) -> typing.Sequence:
-        """ Return the items of the list. Thread safe. """
-        return getattr(self.source, self.__key)
-
-
 class PropertyBinding(Binding):
     """Two way binding from a source property to a target.
 
@@ -230,28 +180,43 @@ class PropertyBinding(Binding):
     # TODO: generalize to 'getter binding'
     # TODO: generalize to 'two way getter connection'
 
-    def __init__(self, source, property_name: str, *, converter=None, validator=None, fallback=None):
+    def __init__(self, source: Observable.ObservableLike, property_name: str, *,
+                 converter: typing.Optional[Converter.ConverterLike[typing.Any, typing.Any]] = None,
+                 validator: typing.Optional[Validator.ValidatorLike[typing.Any]] = None,
+                 fallback: typing.Optional[typing.Any] = None) -> None:
         super().__init__(source, converter=converter, validator=validator, fallback=fallback)
         self.__property_name = property_name
 
-        # thread safe
-        def property_changed(property_name_: str) -> None:
-            assert not self._closed
-            if property_name_ == self.__property_name:
-                value = self.source_getter()
+        # thread safe. careful not to have reference to self, otherwise binding can't be garbage collected.
+        def property_changed(binding: typing.Optional[PropertyBinding], property_name_: str) -> None:
+            assert binding and not binding._closed
+            if property_name_ == binding.property_name:
+                assert callable(binding.source_getter)
+                value = binding.source_getter()
                 if value is not None:
-                    self.update_target(value)
+                    binding.update_target(value)
                 else:
-                    self.update_target_direct(self.fallback)
+                    binding.update_target_direct(binding.fallback)
 
-        self.__property_changed_listener = source.property_changed_event.listen(property_changed)
-        self.source_setter = lambda value: setattr(self.source, self.__property_name, value)
-        self.source_getter = lambda: getattr(self.source, self.__property_name)
+        self.__property_changed_listener = source.property_changed_event.listen(weak_partial(property_changed, self))
 
-    def close(self):
-        self.__property_changed_listener.close()
-        self.__property_changed_listener = None
-        super().close()
+        def set_property_value(source: typing.Any, value: typing.Any) -> None:
+            try:
+                if source:
+                    setattr(source, property_name, value)
+            except AttributeError as exc:
+                raise AttributeError(property_name) from None
+
+        def get_property_value(source: typing.Any) -> typing.Any:
+            return getattr(source, property_name) if source else None
+
+        # configure setting/getter, being careful not to hold references to self
+        self.source_setter = weak_partial(set_property_value, self.source)
+        self.source_getter = weak_partial(get_property_value, self.source)
+
+    @property
+    def property_name(self) -> str:
+        return self.__property_name
 
 
 class PropertyAttributeBinding(Binding):
@@ -272,42 +237,41 @@ class PropertyAttributeBinding(Binding):
     The owner should call close on this object.
     """
 
-    def __init__(self, source, property_name: str, attribute_name: str, converter=None, fallback=None, update_attribute_fn=None):
+    def __init__(self, source: Observable.Observable, property_name: str, attribute_name: str, *,
+                 converter: typing.Optional[Converter.ConverterLike[typing.Any, typing.Any]] = None,
+                 fallback: typing.Optional[typing.Any] = None,
+                 update_attribute_fn: typing.Optional[typing.Callable[[typing.Any, str, typing.Any], typing.Any]] = None) -> None:
         super().__init__(source, converter=converter, fallback=fallback)
         self.__property_name = property_name
         self.__attribute_name = attribute_name
 
         # thread safe
-        def property_changed(property_name_: str):
-            if property_name_ == self.__property_name:
+        def property_changed(binding: typing.Optional[PropertyAttributeBinding], property_name_: str) -> None:
+            if binding and property_name_ == property_name:
                 # perform on the main thread
                 value = getattr(source, property_name)
-                if value is not None and hasattr(value, self.__attribute_name):
-                    self.update_target(getattr(value, self.__attribute_name))
+                if value is not None and hasattr(value, attribute_name):
+                    binding.update_target(getattr(value, attribute_name))
                 else:
-                    self.update_target_direct(self.fallback)
+                    binding.update_target_direct(binding.fallback)
 
-        self.__property_changed_listener = source.property_changed_event.listen(property_changed)
+        self.__property_changed_listener = source.property_changed_event.listen(weak_partial(property_changed, self))
 
-        def source_setter(value):  # pylint: disable=missing-docstring
-            source_value = getattr(self.source, self.__property_name)
+        def source_setter(source: typing.Any, value: typing.Any) -> None:
+            source_value = getattr(source, property_name) if source else None
             if callable(update_attribute_fn):
-                source_value = update_attribute_fn(source_value, self.__attribute_name, value)
+                source_value = update_attribute_fn(source_value, attribute_name, value)
             else:
-                setattr(source_value, self.__attribute_name, value)
-            setattr(self.source, self.__property_name, source_value)
+                setattr(source_value, attribute_name, value)
+            setattr(source, property_name, source_value)
 
-        def source_getter():  # pylint: disable=missing-docstring
-            source_value = getattr(self.source, self.__property_name)
-            return getattr(source_value, self.__attribute_name) if source_value is not None and hasattr(source_value, self.__attribute_name) else None
+        def source_getter(source: typing.Any) -> typing.Any:
+            source_value = getattr(source, property_name) if source else None
+            return getattr(source_value, attribute_name) if source_value is not None and hasattr(source_value, attribute_name) else None
 
-        self.source_setter = source_setter
-        self.source_getter = source_getter
-
-    def close(self):
-        self.__property_changed_listener.close()
-        self.__property_changed_listener = None
-        super().close()
+        # configure setting/getter, being careful not to hold references to self
+        self.source_setter = weak_partial(source_setter, self.source)
+        self.source_getter = weak_partial(source_getter, self.source)
 
 
 class TuplePropertyBinding(Binding):
@@ -328,39 +292,35 @@ class TuplePropertyBinding(Binding):
     The owner should call close on this object.
     """
 
-    def __init__(self, source, property_name: str, tuple_index: int, converter=None, fallback=None):
+    def __init__(self, source: Observable.Observable, property_name: str, tuple_index: int, *,
+                 converter: typing.Optional[Converter.ConverterLike[typing.Any, typing.Any]] = None,
+                 fallback: typing.Optional[typing.Any] = None) -> None:
         super().__init__(source, converter=converter, fallback=fallback)
         self.__property_name = property_name
-        self.__tuple_index = tuple_index
 
         # thread safe
-        def property_changed(property_name_: str):
-            if property_name_ == self.__property_name:
+        def property_changed(binding: TuplePropertyBinding, property_name_: str) -> None:
+            if property_name_ == property_name:
                 # perform on the main thread
                 value = getattr(source, property_name)
-                if value is not None and self.__tuple_index < len(value):
-                    self.update_target(value[self.__tuple_index])
+                if value is not None and tuple_index < len(value):
+                    binding.update_target(value[tuple_index])
                 else:
-                    self.update_target_direct(self.fallback)
+                    binding.update_target_direct(binding.fallback)
 
-        self.__property_changed_listener = source.property_changed_event.listen(property_changed)
+        self.__property_changed_listener = source.property_changed_event.listen(weak_partial(property_changed, self))
 
-        def source_setter(value):  # pylint: disable=missing-docstring
-            source_tuple = getattr(self.source, self.__property_name)
+        def source_setter(source: typing.Any, value: typing.Any) -> None:
+            source_tuple = getattr(source, property_name)
             tuple_as_list = list(source_tuple) if source_tuple is not None else list()
-            while len(tuple_as_list) <= self.__tuple_index:
+            while len(tuple_as_list) <= tuple_index:
                 tuple_as_list.append(None)
-            tuple_as_list[self.__tuple_index] = value
-            setattr(self.source, self.__property_name, tuple(tuple_as_list))
+            tuple_as_list[tuple_index] = value
+            setattr(source, property_name, tuple(tuple_as_list))
 
-        def source_getter():  # pylint: disable=missing-docstring
-            tuple_value = getattr(self.source, self.__property_name)
-            return tuple_value[self.__tuple_index] if tuple_value is not None and self.__tuple_index < len(tuple_value) else None
+        def source_getter(source: typing.Any) -> typing.Any:
+            tuple_value = getattr(source, property_name)
+            return tuple_value[tuple_index] if tuple_value is not None and tuple_index < len(tuple_value) else None
 
-        self.source_setter = source_setter
-        self.source_getter = source_getter
-
-    def close(self):
-        self.__property_changed_listener.close()
-        self.__property_changed_listener = None
-        super().close()
+        self.source_setter = weak_partial(source_setter, self.source)
+        self.source_getter = weak_partial(source_getter, self.source)
