@@ -8,24 +8,24 @@ from nion.utils import Selection
 
 class TestSelection(unittest.TestCase):
 
-    def setUp(self):
+    def setUp(self) -> None:
         pass
 
-    def tearDown(self):
+    def tearDown(self) -> None:
         pass
 
-    def test_extending_selection_with_no_anchor_behaves_sensibly(self):
+    def test_extending_selection_with_no_anchor_behaves_sensibly(self) -> None:
         s = Selection.IndexedSelection()
         s.extend(0)
         self.assertEqual({0}, s.indexes)
         self.assertEqual(0, s.anchor_index)
 
-    def test_inserting_index_into_empty_selection_results_in_empty_index(self):
+    def test_inserting_index_into_empty_selection_results_in_empty_index(self) -> None:
         s = Selection.IndexedSelection()
         s.insert_index(0)
         self.assertEqual(set(), s.indexes)
 
-    def test_inserting_index_into_non_empty_selection_adjusts_indexes(self):
+    def test_inserting_index_into_non_empty_selection_adjusts_indexes(self) -> None:
         s = Selection.IndexedSelection()
         s.set_multiple({0, 2, 4})
         s.insert_index(1)
